from mxcubeweb.core.adapter.adapter_base import AdapterBase


class DiffractometerAdapter(AdapterBase):
    ATTRIBUTES = ["head_configuration"]
    METHODS = ["set_chip_layout"]

    def __init__(self, ho, *args):
        """
        Args:
            (object): Hardware object.
        """
        super().__init__(ho, *args)
        ho.connect("stateChanged", self._state_change)
        ho.connect("valueChanged", self._state_change)
        ho.connect("phaseChanged", self._diffractometer_phase_changed)

    def _diffractometer_phase_changed(self, phase):
        self.app.server.emit(
            "diff_phase_changed",
            {"msg": "Diffractometer phase changed", "phase": phase},
            namespace="/hwr",
        )

    def _state_change(self, *args, **kwargs):
        self.state_change(*args, **kwargs)

    def stop(self):
        pass

    def state(self):
        return "READY" if self._ho.is_ready() else "BUSY"

    def head_configuration(self) -> dict:
        data = self._ho.get_head_configuration()
        data = data.dict() if data else {}

        return data

    def set_chip_layout(
        self,
        layout_name: str,
    ) -> bool:
        self._ho.set_chip_layout(layout_name)
        return True



import logging
from os.path import (
    isfile,
    join,
)
from subprocess import check_output

from flask import (
    Blueprint,
    Response,
    jsonify,
    render_template,
    request,
)
from mxcubecore.model import queue_model_objects as qmo

from . import signals


# Disabling C901 function is too complex (19)
def init_route(app, server, url_prefix):  # noqa: C901
    bp = Blueprint("lims", __name__, url_prefix=url_prefix)

    @bp.route("/synch_samples", methods=["POST"])
    @server.restrict
    def proposal_samples():
        try:
            lims_name = request.get_json().get("lims", None)
            res = jsonify(app.lims.synch_with_lims(lims_name))
        except Exception:
            logging.getLogger("MX3.HWR").exception("Could not synchronize with Lims")
            res = (
                "Could not synchronize with LIMS",
                409,
                {
                    "Content-Type": "application/json",
                },
            )

        return res

    @bp.route("/proposal", methods=["POST"])
    @server.restrict
    def set_proposal():
        """
        Set the selected proposal.
        """
        # proposal_number is the session identifier
        session_id = request.get_json().get("proposal_number", None)
        app.lims.select_session(session_id)
        app.usermanager.update_active_users()

        return Response(status=200)

    @bp.route("/proposal", methods=["GET"])
    @server.restrict
    def get_proposal():
        """
        Return the currently selected proposal. (The proposal list is part of the login_res)
        """
        return jsonify({"Proposal": app.lims.get_proposal_info()})

    def run_get_result_script(script_name, url):
        return check_output(["node", script_name, url], close_fds=True)

    def result_file_test(prefix):
        return isfile(join(server.flask.template_folder, prefix))

    def apply_template(name, data):
        try:
            r = jsonify({"result": render_template(name, data=data)})
        except Exception:
            r = jsonify({"result": "No results yet, processing ..."})

        return r

    @bp.route("/results", methods=["POST"])
    @server.restrict
    def get_results():
        """ """
        qid = request.get_json().get("qid", None)
        r = jsonify({"result": ""})

        if qid:
            model, entry = app.queue.get_entry(qid)
            data = app.queue.queue_to_dict([model], True)
            signals.update_task_result(entry)

            if isinstance(model, qmo.DataCollection):
                if result_file_test("data-collection-results.js"):
                    pass
                elif result_file_test("data-collection-results.html"):
                    r = apply_template("data-collection-results.html", data)

            elif isinstance(model, qmo.Characterisation) or isinstance(
                model, qmo.Workflow
            ):
                if result_file_test("characterisation-results.js"):
                    try:
                        url_list = data["limsResultData"]["workflow_result_url_list"]
                    except Exception as ex:
                        logging.getLogger("MX3.HWR").warning(
                            "Error retrieving wf url list, {0}".format(ex.message)
                        )
                        url_list = None

                    if url_list:
                        r = jsonify(
                            {
                                "result": run_get_result_script(
                                    join(
                                        server.template_folder,
                                        "characterisation-results.js",
                                    ),
                                    url_list[0],
                                )
                            }
                        )
                    else:
                        r = apply_template("data-collection-results.html", data)

                elif result_file_test("characterisation-results.html"):
                    r = apply_template("characterisation-results.html", data)

            elif isinstance(model, qmo.Workflow):
                pass
            elif isinstance(model, qmo.XRFSpectrum):
                pass
            elif isinstance(model, qmo.EnergyScan):
                pass
            else:
                pass

        return r

    return bp



from mxcubeweb.core.adapter.adapter_base import ActuatorAdapterBase
from mxcubeweb.core.models.adaptermodels import (
    HOBeamModel,
    HOBeamValueModel,
)
from mxcubeweb.core.util.adapterutils import export


class BeamAdapter(ActuatorAdapterBase):
    def __init__(self, ho, *args):
        super().__init__(ho, *args)

    def limits(self):
        return -1, -1

    def _get_aperture(self) -> tuple:
        """
        Returns list of apertures and the one currently used.

        :return: Tuple, (list of apertures, current aperture)
        :rtype: tuple
        """
        beam_ho = self._ho

        aperture_list = beam_ho.get_available_size()["values"]
        current_aperture = beam_ho.get_value()[-1]

        return aperture_list, current_aperture

    def _get_value(self) -> HOBeamValueModel:
        beam_ho = self._ho

        beam_info_dict = {
            "position": [],
            "shape": "",
            "size_x": 0,
            "size_y": 0,
        }
        sx, sy, shape, _label = beam_ho.get_value()

        if beam_ho is not None:
            beam_info_dict.update(
                {
                    "position": beam_ho.get_beam_position_on_screen(),
                    "size_x": sx,
                    "size_y": sy,
                    "shape": shape.value,
                }
            )

        aperture_list, current_aperture = self._get_aperture()

        beam_info_dict.update(
            {
                "apertureList": aperture_list,
                "currentAperture": current_aperture,
            }
        )

        return HOBeamValueModel(value=beam_info_dict)

    @export
    def get_size(self) -> HOBeamModel:
        pass

    @export
    def set_size(self, value: HOBeamModel) -> HOBeamModel:
        pass

    def data(self) -> HOBeamModel:
        return HOBeamModel(**self._dict_repr())



import logging

from mxcubecore import HardwareRepository as HWR

from mxcubeweb.core.adapter.beamline_adapter import BeamlineAdapter
from mxcubeweb.core.components.component_base import ComponentBase
from mxcubeweb.core.components.queue import READY


class Beamline(ComponentBase):
    def __init__(self, app, config):
        super().__init__(app, config)

    def init_signals(self):
        from mxcubeweb.routes import signals

        try:
            beamInfo = HWR.beamline.beam
            if beamInfo is not None:
                for sig in signals.beam_signals:
                    beamInfo.connect(beamInfo, sig, signals.beam_changed)
            else:
                logging.getLogger("MX3.HWR").error("beam_info is not defined")
        except Exception:
            msg = "error connecting to beamline_adapter/beam_info hardware object "
            msg += "signals"
            logging.getLogger("MX3.HWR").exception(msg)
        try:
            actions = HWR.beamline.beamline_actions
            if actions is not None:
                cmds = (
                    HWR.beamline.beamline_actions.get_commands()
                    + HWR.beamline.beamline_actions.get_annotated_commands()
                )
                for cmd in cmds:
                    cmd.connect(
                        "commandBeginWaitReply",
                        signals.beamline_action_start,
                    )
                    cmd.connect(
                        "commandReplyArrived",
                        signals.beamline_action_done,
                    )
                    cmd.connect(
                        "commandFailed",
                        signals.beamline_action_failed,
                    )
            else:
                logging.getLogger("MX3.HWR").error(
                    "beamline_actions hardware object is not defined"
                )
        except Exception:
            msg = "error connecting to beamline actions hardware object signals"
            logging.getLogger("MX3.HWR").exception(msg)

        if HWR.beamline.xrf_spectrum:
            HWR.beamline.xrf_spectrum.connect(
                HWR.beamline.xrf_spectrum,
                "xrf_task_progress",
                signals.xrf_task_progress,
            )

    def diffractometer_init_signals(self):
        """
        Connect all the relevant hwobj signals with the corresponding
        callback method.
        """
        from mxcubeweb.routes import signals

        diffractometer = HWR.beamline.diffractometer
        diffractometer.connect("phaseChanged", signals.diffractometer_phase_changed)

    def get_aperture(self):
        """
        Returns list of apertures and the one currently used.

        :return: Tuple, (list of apertures, current aperture)
        :rtype: tuple
        """
        beam = HWR.beamline.beam

        aperture_list = beam.get_available_size()["values"]
        current_aperture = beam.get_value()[-1]

        return aperture_list, current_aperture

    def get_viewport_info(self):
        """
        Get information about current "view port" video dimension, beam position,
        pixels per mm, returns a dictionary with the format:

            data = {"pixelsPerMm": pixelsPerMm,
                    "imageWidth": width,
                    "imageHeight": height,
                    "format": fmt,
                    "sourceIsScalable": source_is_scalable,
                    "scale": scale,
                    "videoSizes": video_sizes,
                    "position": position,
                    "shape": shape,
                    "size_x": sx, "size_y": sy}

        :returns: Dictionary with view port data, with format described above
        :rtype: dict
        """
        fmt, source_is_scalable = "MJPEG", False

        if self.app.CONFIG.app.VIDEO_FORMAT == "MPEG1":
            fmt, source_is_scalable = "MPEG1", True
            video_sizes = HWR.beamline.sample_view.camera.get_available_stream_sizes()
            (width, height, scale) = HWR.beamline.sample_view.camera.get_stream_size()
        else:
            scale = 1
            width = HWR.beamline.sample_view.camera.get_width()
            height = HWR.beamline.sample_view.camera.get_height()
            video_sizes = [(width, height)]

        pixelsPerMm = HWR.beamline.diffractometer.get_pixels_per_mm()

        beam_info_dict = self.get_beam_info()

        data = {
            "pixelsPerMm": pixelsPerMm,
            "imageWidth": width,
            "imageHeight": height,
            "format": fmt,
            "sourceIsScalable": source_is_scalable,
            "scale": scale,
            "videoSizes": video_sizes,
            "videoHash": HWR.beamline.sample_view.camera.stream_hash,
            "videoURL": self.app.CONFIG.app.VIDEO_STREAM_URL,
        }

        data.update(beam_info_dict)
        return data

    def beamline_get_all_attributes(self):
        ho = BeamlineAdapter(HWR.beamline)
        data = ho.dict()
        actions = list()

        try:
            cmds = HWR.beamline.beamline_actions.get_commands()
        except Exception:
            cmds = []
        for cmd in cmds:
            args = []
            for arg in cmd.get_arguments():
                argname = arg[0]
                argtype = arg[1]
                args.append({"name": argname, "type": argtype})
                if argtype == "combo":
                    args[-1]["items"] = cmd.get_combo_argument_items(argname)

            actions.append(
                {
                    "name": cmd.name(),
                    "username": cmd.name(),
                    "state": READY,
                    "arguments": args,
                    "argument_type": cmd.argument_type,
                    "messages": [],
                    "type": cmd.type,
                    "data": cmd.value(),
                }
            )

        actions.extend(self.beamline_get_actions())

        data.update({"availableMethods": ho.get_available_methods()})

        data.update(
            {
                "path": HWR.beamline.session.get_base_image_directory(),
                "actionsList": actions,
            }
        )

        data.update(
            {"energyScanElements": ho.get_available_elements().get("elements", [])}
        )

        data.update(self.diffractometer_get_info())

        return data

    def beamline_get_actions(self):
        actions = []
        beamline_actions = HWR.beamline.beamline_actions

        if getattr(beamline_actions, "pydantic_model", None):
            for cmd_name in beamline_actions.exported_attributes.keys():
                cmd_object = beamline_actions.get_annotated_command(cmd_name)

                actions.append(
                    {
                        "name": cmd_name,
                        "username": cmd_object.name(),
                        "state": READY,
                        "arguments": beamline_actions.exported_attributes[cmd_name][
                            "signature"
                        ],
                        "argument_type": "JSONSchema",
                        "schema": beamline_actions.exported_attributes[cmd_name][
                            "schema"
                        ],
                        "messages": [],
                        "type": "JSONSchema",
                        "data": "",
                    }
                )

        return actions

    def beamline_abort_action(self, name):
        """
        Aborts an action in progress.

        :param str name: Owner / Actuator of the process/action to abort

        """
        beamline_action_names = [
            cmd.name() for cmd in HWR.beamline.beamline_actions.get_commands()
        ]

        if name in beamline_action_names:
            HWR.beamline.beamline_actions.abort_command(name)
        else:
            try:
                ho = HWR.beamline.get_hardware_object(name.lower())
            except AttributeError:
                pass
            else:
                ho.stop()

    def beamline_run_action(self, name, params):
        """
        Starts beamline action with name <name> and passes params as arguments

        : param str name: action to run
        """
        try:
            HWR.beamline.beamline_actions.execute_command(name, params)
        except Exception as ex:
            msg = "Action cannot run: command '%s' does not exist" % name
            raise Exception(msg) from ex

    def get_beam_info(self):
        """
        Returns beam information retrieved by the beam_info hardware object,
        containing position, size and shape.

        :return: Beam info dictionary with keys: position, shape, size_x, size_y
        :rtype: dict
        """
        beam = HWR.beamline.beam
        beam_info_dict = {
            "position": [],
            "shape": "",
            "size_x": 0,
            "size_y": 0,
        }
        sx, sy, shape, _label = beam.get_value()

        if beam is not None:
            beam_info_dict.update(
                {
                    "position": beam.get_beam_position_on_screen(),
                    "size_x": sx,
                    "size_y": sy,
                    "shape": shape.value,
                }
            )

        aperture_list, current_aperture = self.get_aperture()

        beam_info_dict.update(
            {
                "apertureList": aperture_list,
                "currentAperture": current_aperture,
            }
        )

        return beam_info_dict

    def prepare_beamline_for_sample(self):
        if hasattr(HWR.beamline.collect, "prepare_for_new_sample"):
            HWR.beamline.collect.prepare_for_new_sample()

    def diffractometer_set_phase(self, phase):
        try:
            HWR.beamline.diffractometer.wait_ready(30)
        except Exception:
            logging.getLogger("MX3.HWR").warning("Diffractometer not ready")

        HWR.beamline.diffractometer.set_phase(phase)

    def set_aperture(self, pos):
        beam = HWR.beamline.beam
        msg = "Changing beam size to: %s" % pos
        logging.getLogger("MX3.HWR").info(msg)
        beam.set_value(pos)

    def diffractometer_get_info(self):
        ret = {}

        try:
            ret["useSC"] = HWR.beamline.diffractometer.use_sc
        except AttributeError:
            ret["useSC"] = False

        try:
            ret["currentPhase"] = HWR.beamline.diffractometer.get_current_phase()
        except AttributeError:
            ret["currentPhase"] = "None"

        try:
            ret["phaseList"] = HWR.beamline.diffractometer.get_phase_list()
        except AttributeError:
            ret["phaseList"] = []

        return ret

    def get_detector_info(self):
        try:
            filetype = HWR.beamline.detector.get_property("file_suffix")
        except Exception:
            filetype = None

        if filetype is None:
            filetype = "cbf"
            logging.getLogger("MX3.HWR").warning(
                "Detector file format not specified. Using cbf."
            )

        return filetype

    def display_image(self, path, img_num):
        res = {"path": "", "img": 0}

        if path:
            fpath, img = HWR.beamline.detector.get_actual_file_path(path, img_num)
            HWR.beamline.collect.adxv_notify(fpath, img)
            fpath = HWR.beamline.session.get_path_with_proposal_as_root(fpath)

            res = {"path": fpath, "img_num": img_num}

        return res



import logging
import typing

from flask import (
    Blueprint,
    Response,
    jsonify,
    make_response,
    request,
)
from markupsafe import escape
from mxcubecore import HardwareRepository as HWR
from werkzeug.exceptions import UnsupportedMediaType


def create_get_route(app, server, bp, adapter, attr):
    atype = adapter.adapter_type.lower()
    model = adapter._model_from_typehint(getattr(adapter, attr, None))

    @server.restrict
    @server.validate(json=model["args"])
    def get_func(name):
        try:
            args = request.get_json()
        except UnsupportedMediaType:
            args = {}
        result = model["return"](
            **{"return": getattr(app.mxcubecore.get_adapter(name), attr)(**args)}
        )

        return make_response(result.dict(), 200)

    route_url = f"{atype}/<string:name>/{attr}"
    endpoint = f"{atype}_{attr}"
    get_func.__name__ = endpoint
    bp.add_url_rule(route_url, view_func=get_func, endpoint=endpoint, methods=["POST"])


def create_set_route(app, server, bp, adapter, attr, name):
    atype = adapter.adapter_type.lower()
    func = getattr(adapter, attr)
    set_type_hint = typing.get_type_hints(func)

    if "value" in set_type_hint:
        route_url = (
            f"{atype}/{name}/<string:name>" if name else f"{atype}/<string:name>"
        )
        endpoint = f"{atype}_set_{name}" if name else f"{atype}_set"

        @bp.route(route_url, endpoint=endpoint, methods=["PUT"])
        @server.require_control
        @server.restrict
        @server.validate(json=set_type_hint["value"])
        def set_func(name, _th=set_type_hint):
            """
            Tries to set < name > to value
            Replies with status code 200 on success and 400 on exceptions.
            """
            rd = _th["value"].parse_raw(request.data)
            try:
                getattr(app.mxcubecore.get_adapter(rd.name.lower()), attr)(rd)
                return "Value set successfully"
            except Exception:
                logging.getLogger("user_level_log").error(
                    "Could not set %s", rd.name.capitalize()
                )
                return make_response(f"Could not set {rd.name.capitalize()}", 400)

        set_func.__name__ = f"{atype}_set_value"


def create_route(app, server, bp, adapter, obj, cmd_name):
    route_url = f"/{obj}/command/{cmd_name}"
    arg_schema = adapter._pydantic_model_for_command(cmd_name)

    @bp.route(route_url, endpoint=cmd_name, methods=["POST"])
    @server.require_control
    @server.restrict
    @server.validate(json=arg_schema)
    def set_func():
        """
        Tries to set < name > to value
        Replies with status code 200 on success and 409 on exceptions.
        """
        args = request.get_json()
        adapter = app.mxcubecore.get_adapter(obj.lower())
        adapter.execute_command(cmd_name, args)
        return "Command executed successfull"

    set_func.__name__ = f"{obj}_{cmd_name}"


def add_adapter_routes(app, server, bp):
    adapter_type_list = []

    for _id, a in app.mxcubecore.adapter_dict.items():
        adapter = a["adapter"]
        # Only add the route once for each type (class) of adapter
        if adapter.adapter_type not in adapter_type_list:
            adapter_type_list.append(adapter.adapter_type)

            # All adapters, inheriting BaseAdapter have _set_value() to
            # set the value of the underlyaing hardware object and
            # data() to return a representation of the object, so we are
            # mapping these by default
            create_set_route(app, server, bp, adapter, "_set_value", "value")
            create_get_route(app, server, bp, adapter, "data")

            # Map all other functions starting with prefix get_ or set_ and
            for attr in dir(adapter):
                if attr.startswith("get"):
                    create_get_route(app, server, bp, adapter, attr)

                if attr.startswith("set"):
                    create_set_route(
                        app,
                        server,
                        bp,
                        adapter,
                        attr,
                        attr.replace("set_", ""),
                    )

        exported_methods = adapter._exported_methods()

        for cmd_name in exported_methods.keys():
            create_route(app, server, bp, adapter, _id, cmd_name)


def init_route(app, server, url_prefix):
    bp = Blueprint("beamline", __name__, url_prefix=url_prefix)

    add_adapter_routes(app, server, bp)

    @bp.route("/", methods=["GET"])
    @server.restrict
    def beamline_get_all_attributes():
        return jsonify(app.beamline.beamline_get_all_attributes())

    @bp.route("/<name>/abort", methods=["GET"])
    @server.require_control
    @server.restrict
    def beamline_abort_action(name):
        """
        Aborts an action in progress.

        :param str name: Owner / Actuator of the process/action to abort

        Replies with status code 200 on success and 500 on exceptions.
        """
        try:
            app.beamline.beamline_abort_action(name)
        except Exception:
            logging.getLogger("MX3.HWR").exception("Could not abort %s", name)
            return make_response(f"Could not abort {escape(name)}", 500)
        else:
            logging.getLogger("user_level_log").error("%s, aborted" % name)
            return make_response("{}", 200)

    @bp.route("/<name>/run", methods=["POST"])
    @server.require_control
    @server.restrict
    def beamline_run_action(name):
        """
        Starts a beamline action; POST payload is a json-encoded object with
        'parameters' as a list of parameters

        :param str name: action to run

        Replies with status code 200 on success and 500 on exceptions.
        """
        try:
            params = request.get_json()["parameters"]
        except Exception:
            params = []

        try:
            app.beamline.beamline_run_action(name, params)
        except Exception:
            logging.getLogger("MX3.HWR").exception("Cannot run action %s", name)
            return make_response(f"Cannot run action {escape(name)}", 500)
        else:
            return make_response("{}", 200)

    @bp.route("/beam/info", methods=["GET"])
    @server.restrict
    def get_beam_info():
        """
        Beam information: position, size, shape
        return_data = {"position": , "shape": , "size_x": , "size_y": }
        """
        return jsonify(app.beamline.get_beam_info())

    @bp.route("/datapath", methods=["GET"])
    @server.restrict
    def beamline_get_data_path():
        """
        Retrieve data directory from the session hwobj,
        this is specific for each beamline.
        """
        data = HWR.beamline.session.get_base_image_directory()
        return jsonify({"path": data})

    @bp.route("/prepare_beamline", methods=["PUT"])
    @server.require_control
    @server.restrict
    def prepare_beamline_for_sample():
        """
        Prepare the beamline for a new sample.
        """
        try:
            app.beamline.prepare_beamline_for_sample()
        except Exception:
            msg = "Cannot prepare the Beamline for a new sample"
            logging.getLogger("HWR").exception(msg)
            return Response(status=200)
        return Response(status=200)

    return bp



import logging
from enum import Enum

from mxcubeweb.core.adapter.adapter_base import ActuatorAdapterBase
from mxcubeweb.core.models.adaptermodels import (
    HOActuatorValueChangeModel,
    NStateModel,
    StrValueModel,
)


class BeamlineActionAdapter(ActuatorAdapterBase):
    def __init__(self, ho, *args, **kwargs):
        """
        Args:
            (object): Hardware object.
        """
        super().__init__(ho, *args, **kwargs)
        self._value_change_model = HOActuatorValueChangeModel

        ho.connect("valueChanged", self._value_change)
        ho.connect("stateChanged", self.state_change)

    def _value_change(self, value):
        if isinstance(value, Enum):
            v = value.name
        else:
            v = value

        self.value_change(v)

    def commands(self):
        method_list = [
            attribute
            for attribute in dir(self._ho.__class__)
            if callable(getattr(self._ho.__class__, attribute))
            and attribute.startswith("_") is False
        ]

        return method_list

    def _set_value(self, value: HOActuatorValueChangeModel):
        self._ho.set_value(self._ho.VALUES[value.value])

    def _get_value(self) -> StrValueModel:
        return StrValueModel(value=self._ho.get_value().name)

    def msg(self):
        try:
            msg = self._ho.get_value().name
        except Exception:
            msg = "---"
            logging.getLogger("MX3.HWR").error(
                "Failed to get beamline attribute message"
            )

        return msg

    def data(self) -> NStateModel:
        return NStateModel(**self._dict_repr())
